"""Exposes constants."""

SQL_INSERT_RUN = """
INSERT INTO tracing_run (uuid, app_name, column_names)  VALUES($1, $2, $3);
"""

SQL_INSERT_ROW = """
INSERT INTO tracing_row (uuid, row_data) VALUES($1, $2);
"""

SQL_SELECT_NUMBER_OF_COLS = """
SELECT CARDINALITY(column_names) AS col_count, uuid 
FROM tracing_run WHERE uuid=$1
"""

SQL_SELECT_COL_NAMES = """
select unnest as col_name, ordinality -1 as index 
from unnest((select column_names from tracing_run where uuid=$1)) 
with ordinality
"""

SQL_SELECT_LATEST_RUN = """
select uuid from tracing_run where app_name=$1 
order by creation_time desc limit 1
"""

SQL_SELECT_ALL_TRACER_RUNS = """
 select app_name  from tracing_run group by app_name order by  max(creation_time) desc
"""

SQL_SELECT_RUNS = """
select creation_time as creation_time, uuid from tracing_run where app_name = $1 order by creation_time desc;
"""


SQL_SELECT_APP_NAME = """
Select app_name, creation_time from tracing_run where uuid=$1
"""

SQL_SELECT_RUN_INFO = """
select                                                                             
    count(*) as counter, 
    min(a.date_time) as from_time,  
    max(a.date_time) as to_time
from 
    tracing_row a
where 
    a.uuid=$1
"""

MINUTE_IN_SECONDS = 60
HOUR_IN_SECONDS = 60 * MINUTE_IN_SECONDS
DAY_IN_SECONDS = 24 * HOUR_IN_SECONDS



"""Dolon exceptions."""


class DolonException(Exception):
    """A generic Dolon exception."""


class InvalidMessage(DolonException):
    """Invalid message."""


class InvalidEnvironmentVariable(DolonException):
    """Invalid environment variable."""



"""Exposes the basic interface to interact with the serialization means."""

import dolon.impl.utils_impl as utils_impl


async def process_message(db, payload):
    """Processes a tracing message storing it to the db.

    :param db: The database object to use.

    :param dict payload: A dict representing the message to store.

    Can be either a tracing run creation in the form of:

        msg = {
            "msg_type": "create_trace_run",
            "app_name": app_name,
            "uuid": identifier,
            "column_names": ["v1", 'v2']
        }

    or for the insertion of a tracing row:

        msg = {
            "msg_type": "row",
            "uuid": identifier,
            "row_data": [12.2, 123.1]
        }

    raises: InvalidMessage
    """
    await utils_impl.process_message(db, payload)


async def get_trace(uuid):
    """Returns all the tracing rows for the passed in uuid.

    :param str uuid: The identifier for the trace run.

    :returns: A list of strings representing a csv view of the tracing run.
    :rtype: list[str]
    """
    return await utils_impl.get_trace(uuid)


async def get_all_tracers():
    """Returns a list with the names of all tracing runs.

    :returns: A list with the names of all tracing runs.
    :rtype: list[str]
    """
    return await utils_impl.get_all_tracers()


async def get_trace_as_json(uuid):
    """Returns all the tracing rows for the passed in uuid as json.

    :param str uuid: The identifier for the trace run.

    :returns: All the tracing rows for the passed in uuid as json.
    :rtype: list[dict]
    """
    return await utils_impl.get_trace_as_json(uuid)


async def get_trace_run_info(uuid):
    """Returns descriptive info for the passed in uuid.

    :param str uuid: The identifier for the trace run.

    :returns: Descriptive info for the passed in uuid.
    :rtype: dict
    """
    return await utils_impl.get_trace_run_info(uuid)


async def get_latest_trace(app_name):
    """Returns the latest trace for the passed in app_name.

    :return: A list of objects.
    """
    return await utils_impl.get_latest_trace(app_name)



"""Builds the library.

To upload to cheese:

sudo pip3 install twine
rm -rf build dist
sudo python3 setup.py sdist bdist_wheel
twine upload  --skip-existing dist/* --verbose
"""
import pathlib

from setuptools import find_packages, setup

_ROOT = pathlib.Path(__file__).parent

with open(str(_ROOT / 'README.rst')) as f:
    readme = f.read()

setup(
    name="dolon",
    description="A performance tracer application.",
    long_description=readme,
    version='0.0.43',
    packages=find_packages(),
    install_requires=[
        "asyncpg",
        "psutil"
    ],
)



"""The trace client to record diagnostics."""

import asyncio
import logging
import os
import tracemalloc

import psutil

import dolon.profiler as profiler
import dolon.impl.db_conn_impl as db_conn_impl
import dolon.impl.db_stats as db_stats
import dolon.impl.trace_client_impl as trace_client_impl

_logger = logging.getLogger(__name__)


async def start_tracer(app_name, frequency, host, port, *diagnostics):
    """Starts a tracer.

    This is where a tracer is created and starts running in the background. It
    creates a new tracing run with the passed in app_name.

    If other tracing runs with the same name already exists then a new run
    will be created based on the start time without affecting them; doing so
    we keep history untouched and we can view any run from the front end based
    on the combination of app name and creation time.

    If an exception will be caught it will be logged and the application will
    stop.

    Note that all the profile-able functions will be added to diagnostics
    automatically.

    :param str app_name: The application name that will tag the trace.

    :param int frequency: The sleep interval between consecutive traces.

    :param str host: The host of the mnemic service.

    :param int port: The port that is listening for messages.

    :param diagnostics: The diagnostic callable to record. There is no limit
        in the number of diagnostic function to pass and they can differ from
        run to run for the same tracer name. The name of the callable will be
        used to store it and show it from mnemic's front end.
    """
    diagnostics = list(diagnostics) + profiler.get_profiling_functions(True)
    try:
        tc = trace_client_impl.TraceClientImpl(
            app_name,
            host,
            port,
            *diagnostics
        )
        async with tc:
            await tc.run(frequency)
    except Exception as ex:
        _logger.exception(ex)
        exit(-1)


class PostgresDiagnostics(db_conn_impl.DbConnectionImpl):
    """Wraps memory diagnostics within an async context manager.

    Must be used as a context manager; its profiler functions can be passed
    as diagnostics to start_tracer.

    :ivar str conn_str: The connection string to the database.
    """

    def __init__(self, conn_str):
        """Initializer.

        :param str conn_str: The connection string to the psql database.
        """
        super().__init__(min_size=1, max_size=1, conn_str=conn_str)

    async def live_msgs(self):
        """Returns the number of live messages in db.

        :returns: The number of live messages in db. Can be used as a profiler
            function.

        :rtype: async callable.
        """
        return await db_stats.live_msgs_in_db.get_value(self)

    async def dead_msgs(self):
        """Returns the number of dead messages in db.

        :returns: The number of dead messages in db. Can be used as a profiler
            function.

        :rtype: async callable.
        """
        return await db_stats.dead_msgs_in_db.get_value(self)

    async def idle(self):
        """Returns the number of idle sessions in db.

        :returns: The number of idle messages in db. Can be used as a profiler
            function.

        :rtype: async callable.
        """
        return await db_stats.idle_in_db.get_value(self)

    async def count_db_connections(self):
        """Returns the number of connections in db.

        :returns: The number of connections in db. Can be used as a profiler
            function.

        :rtype: async callable.
        """
        return await db_stats.conn_count_in_db.get_value(self)


async def mem_allocation():
    """Returns the size of the allocated memory in Mb.

    Assumes that the client code has already called tracemalloc.start().

    :returns: The memory allocation in Mb. Can be used as a profiler
            function.

    :rtype: int.
    """
    current, peak = tracemalloc.get_traced_memory()
    return current / 10 ** 6


async def active_tasks():
    """Returns the number of active async tasks.

    :returns: The number of active async tasks.

    :rtype: int.
    """
    return len([task for task in asyncio.Task.all_tasks() if not task.done()])


async def cpu_percent():
    """Returns the CPU usage percent.

    :returns: The CPU usage percent.

    :rtype: float.
    """
    return psutil.cpu_percent()


async def virtual_memory_percent():
    """Returns the virtual memory percentage usage.

    :returns: The virtual memory percentage usage.

    :rtype: float.
    """
    return psutil.virtual_memory().percent


async def memory_use():
    """Returns the memory usage.

    :returns: The memory usage.

    :rtype: float.
    """
    pid = os.getpid()
    py = psutil.Process(pid)
    return py.memory_info()[0] / 2. ** 30
